class IPyTVException(Exception):
    pass


class MalformedExtinfException(IPyTVException):
    pass


class MalformedPlaylistException(IPyTVException):
    pass


class URLException(IPyTVException):
    pass


class WrongTypeException(IPyTVException):
    pass


class IndexOutOfBoundsException(IPyTVException):
    pass


class AttributeAlreadyPresentException(IPyTVException):
    pass


class AttributeNotFoundException(IPyTVException):
    pass



import re
from typing import Optional, Dict

M3U_HEADER_TAG = "#EXTM3U"
M3U_EXTINF_REGEX = r'^#EXTINF:[-0-9\.]+,.*$'
M3U_PLUS_EXTINF_REGEX = r'^#EXTINF:[-0-9\.]+(\s+[\w-]+="[^"]*")+,.*$'
M3U_PLUS_EXTINF_PARSE_REGEX = r'^#EXTINF:(?P<duration_g>[-0-9\.]+)' \
    r'(?P<attributes_g>(\s+[\w-]+="[^"]*")*),' \
    r'(?P<name_g>.*)'
M3U_PLUS_BROKEN_EXTINF_PARSE_REGEX = r'^#EXTINF:(?P<duration_g>[-0-9\.]+)' \
    r'(?P<attributes_g>(\s+[\w-]+=".*)*),' \
    r'(?P<name_g>.*)'
M3U_PLUS_BROKEN_ATTRIBUTE_PARSE_REGEX = r'(?:\s+)[\w-]+="'


def is_m3u_header_row(row: str) -> bool:
    return row.startswith(M3U_HEADER_TAG)


def is_m3u_extinf_row(row: str) -> bool:
    return re.search(M3U_EXTINF_REGEX, row) is not None


def is_m3u_plus_extinf_row(row: str) -> bool:
    return re.search(M3U_PLUS_EXTINF_REGEX, row) is not None


def match_m3u_plus_broken_extinf_row(row: str) -> Optional[re.Match]:
    return re.search(M3U_PLUS_BROKEN_EXTINF_PARSE_REGEX, row)


def get_m3u_plus_broken_attributes(row: str) -> Dict[str, str]:
    """In the case of an EXTINF row with a list of "broken" attributes (i.e.
    attributes the value of which contains badly nested dohble quotes), this
    function can parse the list of attributes by:
    1. extract only the attribute list from the whole row (please note: the
    leading space is kept, but the trailing comma is removed);
    2. find all occurrences of the attribute-name=" pattern and extract the
    attribute name from the results;
    3. loop through all the attribute patterns and split the whole row in two
    using the attribute name as separator, and take the right part of the split.
    4. Split the right part in two, by using next attribute as separator: the
    wanted value will be found in the left part of the split.
    5. Replace all double quotes with underscores.
    """
    match = match_m3u_plus_broken_extinf_row(row)
    if match is None:
        return {}
    attributes = match.group("attributes_g").rstrip(',')
    tokens = re.findall(M3U_PLUS_BROKEN_ATTRIBUTE_PARSE_REGEX, attributes)
    attrs = {}
    for i, token in enumerate(tokens):
        name = token.lstrip().rstrip('="')
        right = row.split(token)[1]
        separator = tokens[i+1] if i < len(tokens)-1 else '",'
        left = right.split(separator)[0].rstrip('"')
        # Let's replace double quotes with underscore
        attrs[name] = left.replace('"', '_')
    return attrs


def match_m3u_plus_extinf_row(row: str) -> Optional[re.Match]:
    return re.match(M3U_PLUS_EXTINF_PARSE_REGEX, row)


def is_extinf_row(row: str) -> bool:
    return row.startswith("#EXTINF")


def is_comment_or_tag_row(row: str) -> bool:
    return row.startswith('#')


def is_url_row(row: str) -> bool:
    return not is_comment_or_tag_row(row)



import logging
import shlex
from enum import Enum
from typing import Dict, List

from ipytv import m3u
from ipytv.exceptions import MalformedExtinfException

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class M3UEntry:
    def __init__(self, url: str, name: str = "", duration: str = "-1"):
        self.url = url
        self.name = name
        self.duration = str(duration)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, M3UEntry) \
               and self.url == other.url \
               and self.name == other.name \
               and self.duration == other.duration

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class IPTVAttr(Enum):
    TVG_ID = "tvg-id"
    TVG_NAME = "tvg-name"
    TVG_LANGUAGE = "tvg-language"
    TVG_LOGO = "tvg-logo"
    TVG_LOGO_SMALL = "tvg-logo-small"
    TVG_COUNTRY = "tvg-country"
    GROUP_TITLE = "group-title"
    PARENT_CODE = "parent-code"
    AUDIO_TRACK = "audio-track"
    TVG_SHIFT = "tvg-shift"
    TVG_REC = "tvg-rec"
    ASPECT_RATIO = "aspect-ratio"
    TVG_CHNO = "tvg-chno"
    RADIO = "radio"
    TVG_URL = "tvg-url"


class IPTVChannel(M3UEntry):

    def __init__(self, url: str = "", name: str = "",
                 duration: str = "-1", attributes: Dict[str, str] = None,
                 extras: List[str] = None):
        super().__init__(url, name, duration)
        self.attributes: Dict[str, str] = attributes if attributes is not None else {}
        self.extras: List[str] = extras if extras is not None else []

    def __eq__(self, other: object) -> bool:
        return isinstance(other, IPTVChannel) \
            and super().__eq__(other) \
            and self.attributes == other.attributes \
            and self.extras == other.extras

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def copy(self) -> 'IPTVChannel':
        return IPTVChannel(
            url=self.url,
            name=self.name,
            duration=self.duration,
            attributes=self.attributes.copy(),  # shallow copy is ok, as we're dealing with primitive types
            extras=self.extras.copy()           # shallow copy is ok, as we're dealing with primitive types
        )

    def parse_extinf_string(self, extinf_string: str) -> None:
        match = m3u.match_m3u_plus_extinf_row(extinf_string)
        if match is not None:
            # Case of a well-formed EXTINF row
            log.info("parsing a well-formed EXTINF row:\n%s", extinf_string)
            self.duration = match.group("duration_g")
            log.info("duration: %s", self.duration)
            attributes = match.group("attributes_g")
            for entry in shlex.split(attributes):
                pair = entry.split("=", 1)
                key = pair[0]
                value = pair[1]
                self.attributes[key] = value
            log.info("attributes: %s", self.attributes)
            self.name = match.group("name_g")
            log.info("name: %s", self.name)
            return

        match = m3u.match_m3u_plus_broken_extinf_row(extinf_string)
        if match is not None:
            # Case of a broken #EXTINF row (with quoting issues)
            log.warning("parsing an EXTINF row with quoting issues:\n%s", extinf_string)
            self.duration = match.group("duration_g")
            log.info("duration: %s", self.duration)
            self.attributes = m3u.get_m3u_plus_broken_attributes(extinf_string)
            log.info("attributes: %s", self.attributes)
            self.name = match.group("name_g")
            log.info("name: %s", self.name)
            return

        # This EXTINF row can't be parsed
        log.error("malformed #EXTINF row: %s", extinf_string)
        raise MalformedExtinfException(f"Malformed #EXTINF string:\n{extinf_string}")

    def __str__(self) -> str:
        attr_str = ''
        if len(self.attributes) > 0:
            for name, value in self.attributes.items():
                attr_str += f'{name}: "{value}", '
        if attr_str.endswith(', '):
            attr_str = attr_str[:-2]
        extras_str = ''
        if len(self.extras) > 0:
            for extra in self.extras:
                extras_str += f'{extra}, '
        if extras_str.endswith(', '):
            extras_str = extras_str[:-2]
        out = f'{{name: "{self.name}", duration: "{self.duration}", '\
              f'url: "{self.url}", attributes: {{{attr_str}}}, extras: [{extras_str}]}}'
        return out

    def _build_m3u_plus_extinf_entry(self) -> str:
        extinf_pattern = '#EXTINF:{}{},{}\n'
        attrs = ''
        for key, value in self.attributes.items():
            attrs += f' {key}="{value}"'
        return extinf_pattern.format(
            self.duration,
            attrs,
            self.name
        )

    def _build_m3u8_extinf_entry(self) -> str:
        extinf_pattern = '#EXTINF:{},{}\n'
        return extinf_pattern.format(
            self.duration,
            self.name
        )

    def _build_extras_entry(self) -> str:
        out = ''
        for extra in self.extras:
            out += f'{extra}\n'
        return out

    def _build_url_entry(self) -> str:
        return f"{self.url}\n"

    def to_m3u_plus_playlist_entry(self) -> str:
        extinf_row = self._build_m3u_plus_extinf_entry()
        extras_rows = self._build_extras_entry()
        url_row = self._build_url_entry()
        return f'{extinf_row}{extras_rows}{url_row}'

    def to_m3u8_playlist_entry(self) -> str:
        extinf_row = self._build_m3u8_extinf_entry()
        url_row = self._build_url_entry()
        return f'{extinf_row}{url_row}'


def from_playlist_entry(entry: List[str]) -> 'IPTVChannel':
    channel = IPTVChannel()
    for row in entry:
        if m3u.is_extinf_row(row):
            try:
                channel.parse_extinf_string(row)
            except MalformedExtinfException:
                log.warning("Skipping the following entry as it contains a malformed #EXTINF row:\n%s", entry)
            log.info("#EXTINF row found")
        elif m3u.is_comment_or_tag_row(row):
            # a comment or a non-supported tag, we add it to extras
            channel.extras.append(row)
            log.warning("commented row or unsupported tag found:\n%s", row)
        else:
            channel.url = row
            log.info("URL row found")
    return channel



import logging
import math
import multiprocessing as mp
from multiprocessing.pool import AsyncResult
from typing import List, Dict

import requests
from requests import RequestException

import ipytv.channel
from ipytv import m3u
from ipytv.channel import IPTVChannel, IPTVAttr
from ipytv.exceptions import MalformedPlaylistException, URLException, \
    WrongTypeException, IndexOutOfBoundsException, \
    AttributeAlreadyPresentException, AttributeNotFoundException
from ipytv.m3u import M3U_HEADER_TAG

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

# The value of __MIN_CHUNK_SIZE cannot be smaller than 2
__MIN_CHUNK_SIZE = 20


class M3UPlaylist:
    NO_GROUP_KEY = '_NO_GROUP_'
    NO_URL_KEY = '_NO_URL_'

    def __init__(self):
        self._channels: List[IPTVChannel] = []
        self._attributes: Dict = {}
        self._iter_index: int = -1

    def length(self):
        return len(self._channels) if self._channels is not None else 0

    def _check_attribute(self, name: str) -> None:
        if name not in self._attributes:
            log.error("the attribute %s does not exist", name)
            raise AttributeNotFoundException(f"the attribute {name} does not exist")

    def get_attribute(self, name: str) -> str:
        self._check_attribute(name)
        return self._attributes[name]

    def get_attributes(self) -> Dict:
        return self._attributes

    def add_attribute(self, name: str, value: str) -> None:
        if name not in self._attributes:
            self._attributes[str(name)] = str(value)
            log.info("attribute added: %s: %s", name, value)
        else:
            log.error(
                "the attribute %s is already present with value %s",
                name,
                self._attributes[name]
            )
            raise AttributeAlreadyPresentException(
                f"the attribute {name} is already present with value {self._attributes[name]}"
            )

    def add_attributes(self, attributes: Dict[str, str]) -> None:
        for k, v in attributes.items():
            self.add_attribute(k, v)

    def update_attribute(self, name: str, value: str) -> None:
        self._check_attribute(name)
        self._attributes[name] = value
        log.info("attribute %s updated to value %s", name, value)

    def remove_attribute(self, name: str) -> str:
        self._check_attribute(name)
        attribute = self._attributes[name]
        del self._attributes[name]
        log.info("attribute %s deleted", name)
        return attribute

    def _check_index(self, index: int) -> None:
        length = self.length()
        if index < 0 or index >= length:
            log.error(
                "the index %s is out of the (0, %s) range",
                str(index),
                str(length)
            )
            raise IndexOutOfBoundsException(f"the index {index} is out of the (0, {length}) range")

    def get_channel(self, index: int) -> IPTVChannel:
        self._check_index(index)
        return self._channels[index]

    def get_channels(self) -> List[IPTVChannel]:
        return self._channels

    def insert_channel(self, index: int, channel: IPTVChannel) -> None:
        self._check_index(index)
        self._channels.insert(index, channel)
        log.info("channel %s inserted in position %s", channel, index)

    def insert_channels(self, index: int, chan_list: List[IPTVChannel]) -> None:
        self._check_index(index)
        for i in range(len(chan_list), 0, -1):
            self.insert_channel(index, chan_list[i])
        log.info("%s channels inserted to the playlist in position %s", len(chan_list), index)

    def append_channel(self, channel: IPTVChannel) -> None:
        self._channels.append(channel)
        log.info("channel added: %s", channel)

    def append_channels(self, chan_list: List[IPTVChannel]) -> None:
        self._channels += chan_list
        log.info("%s channels appended to the playlist", len(chan_list))

    def update_channel(self, index: int, channel: IPTVChannel) -> None:
        self._check_index(index)
        self._channels[index] = channel
        log.info("index %s has been updated with channel %s", str(index), channel)

    def remove_channel(self, index: int) -> IPTVChannel:
        self._check_index(index)
        channel = self._channels[index]
        del self._channels[index]
        log.info("the channel with index %s has been deleted", str(index))
        return channel

    def _build_header(self) -> str:
        out = M3U_HEADER_TAG
        for k, v in self._attributes.items():
            out += f' {k}="{v}"'
        return out

    def group_by_attribute(self, attribute: str = IPTVAttr.GROUP_TITLE.value,
                           include_no_group: bool = True) -> Dict:
        groups: Dict[str, List] = {}
        for i, chan in enumerate(self._channels):
            group = self.NO_GROUP_KEY
            if attribute in chan.attributes and len(chan.attributes[attribute]) > 0:
                group = chan.attributes[attribute]
            elif not include_no_group:
                continue
            groups.setdefault(group, [])
            groups[group].append(i)
        return groups

    def group_by_url(self, include_no_group: bool = True) -> Dict:
        groups: Dict[str, List] = {}
        for i, chan in enumerate(self._channels):
            group = self.NO_URL_KEY
            if len(chan.url) > 0:
                group = chan.url
            elif not include_no_group:
                continue
            groups.setdefault(group, [])
            groups[group].append(i)
        return groups

    def to_m3u_plus_playlist(self) -> str:
        out = f"{self._build_header()}\n"
        for channel in self._channels:
            out += channel.to_m3u_plus_playlist_entry()
        return out.rstrip()

    def to_m3u8_playlist(self) -> str:
        out = f"{m3u.M3U_HEADER_TAG}\n"
        for channel in self._channels:
            out += channel.to_m3u8_playlist_entry()
        return out.rstrip()

    def copy(self) -> 'M3UPlaylist':
        new_pl = M3UPlaylist()
        for channel in self._channels:
            new_pl.append_channel(channel.copy())
        new_pl.add_attributes(
            self.get_attributes().copy()     # shallow copy is ok, as we're dealing with primitive types
        )
        return new_pl

    def __eq__(self, other: object) -> bool:
        length = self.length()
        if not isinstance(other, M3UPlaylist) or \
                other.length() != length:
            return False
        if not other.get_attributes() == self.get_attributes():
            return False
        for i, ch in enumerate(self):
            if not other.get_channel(i) == ch:
                return False
        return True

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __str__(self) -> str:
        out = f"attributes: {self._attributes}\n" if len(self._attributes) > 0 else ''
        index = 0
        for chan in self._channels:
            out += f"{index}: {chan}\n"
            index += 1
        return out

    def __iter__(self) -> 'M3UPlaylist':
        self._iter_index = 0
        return self

    def __next__(self) -> IPTVChannel:
        if self._iter_index >= self.length():
            raise StopIteration
        next_chan = self.get_channel(self._iter_index)
        self._iter_index += 1
        return next_chan


def loada(array: List) -> 'M3UPlaylist':
    if not isinstance(array, list):
        log.error("expected %s, got %s", type([]), type(array))
        raise WrongTypeException("Wrong type: array (List) expected")
    if len(array) < 2:
        log.error("a playlist should have at least 2 rows (found %s)", len(array))
        raise MalformedPlaylistException(f"a playlist should have at least 2 rows (found {len(array)})")
    header = array[0].strip()
    body = array[1:]
    if not m3u.is_m3u_header_row(header):
        log.error(
            "the playlist's first row should start with \"%s\", but it's \"%s\"",
            M3U_HEADER_TAG,
            header
        )
        raise MalformedPlaylistException(f"Missing or misplaced {M3U_HEADER_TAG} row")
    out_pl = M3UPlaylist()
    out_pl.add_attributes(_parse_header(header))
    cores = mp.cpu_count()
    log.debug("%s cores detected", cores)
    chunks = _chunk_body(body, cores)
    results: List[AsyncResult] = []
    log.debug("spawning a pool of processes (one per core) to parse the playlist")
    with mp.Pool(processes=cores) as pool:
        for chunk in chunks:
            begin = chunk["begin"]
            end = chunk["end"]
            log.debug(
                "assigning a \"populate\" task (begin: %s, end: %s) to a process in the pool",
                begin,
                end
            )
            result = pool.apply_async(_populate, (body, begin, end))
            results.append(result)
        log.debug("closing workers")
        pool.close()
        log.debug("workers closed")
        log.debug("waiting for workers termination")
        pool.join()
        log.debug("workers terminated")
        for result in results:
            p_list = result.get()
            out_pl.append_channels(p_list.get_channels())
    return out_pl


def loads(string: str) -> 'M3UPlaylist':
    if isinstance(string, str):
        return loada(string.split("\n"))
    log.error("expected %s, got %s", type(''), type(string))
    raise WrongTypeException("Wrong type: string expected")


def loadf(filename: str) -> 'M3UPlaylist':
    if not isinstance(filename, str):
        log.error("expected %s, got %s", type(''), type(filename))
        raise WrongTypeException("Wrong type: string expected")
    with open(filename, encoding='utf-8') as file:
        buffer = file.readlines()
        return loada(buffer)


def loadu(url: str) -> 'M3UPlaylist':
    if not isinstance(url, str):
        log.error("expected %s, got %s", type(''), type(url))
        raise WrongTypeException("Wrong type: string expected")
    try:
        response = requests.get(url, timeout=10)
        if response.ok:
            return loads(response.text)
        raise URLException(
            f"Failure while opening {url}.\nResponse status code: {response.status_code}"
        )
    except RequestException as exception:
        log.error(
            "failure while opening %s: %s",
            url,
            exception
      )
        raise URLException(
            f"Failure while opening {url}.\nError: {exception}"
        ) from exception


def _parse_header(header: str) -> Dict[str, str]:
    attrs = header.replace(f'{M3U_HEADER_TAG}', '').lstrip()
    attributes = {}
    for attr in attrs.split():
        entry = attr.split("=")
        if len(entry) == 2:
            name = entry[0].replace('"', '')
            value = entry[1].replace('"', '')
            attributes[name] = value
    return attributes


def _build_chunk(begin: int, end: int) -> Dict[str, int]:
    return {
        "begin": begin,
        "end": end
    }


def _find_chunk_end(sub_array: List[str]) -> int:
    offset = len(sub_array)
    for offset, row in enumerate(sub_array):
        if m3u.is_extinf_row(row):
            log.debug(
                "chunking at the following row (offset %s) as it's an #EXTINF row:\n%s",
                offset,
                row
            )
            break
        if offset > 0 and m3u.is_url_row(row) and m3u.is_url_row(sub_array[offset-1]):
            log.debug(
                "chunking at the following row (offset %s) as it's a url row after another url row:\n%s",
                offset,
                row
            )
            break
    return offset


def _compute_chunk(array: List, start: int, min_size: int) -> Dict[str, int]:
    length = len(array)
    sub_array = array[start + min_size:]
    if length - start > min_size:
        log.debug(
            "there are enough remaining rows (%s left) to populate at least one full-size chunk",
            length - start
        )
        offset = _find_chunk_end(sub_array)
        end = start + min_size + offset
        log.debug("chunk end found at row %s", end)
        return _build_chunk(start, end)
    log.debug(
        "there are less than (or exactly) %s rows (%s left), so the chunk end is the array end",
        min_size,
        length - start
    )
    return _build_chunk(start, length)


def _chunk_body(array: List, chunk_count: int, enforce_min_size: bool = True) -> List:
    length = len(array)
    chunk_size = math.floor(length / chunk_count)
    if enforce_min_size and chunk_size < __MIN_CHUNK_SIZE:
        log.debug(
            "no chunking as each of the %s chunks would be smaller than the configured minimum (%s < %s)",
            chunk_count,
            chunk_size,
            __MIN_CHUNK_SIZE
        )
        return [
            {
                "begin": 0,
                "end": length
            }
        ]
    chunk_list = []
    start = 0
    while start < length:
        chunk: Dict[str, int] = _compute_chunk(array, start, chunk_size)
        chunk_list.append(chunk)
        start = chunk["end"]
    log.debug("chunk_list: %s", chunk_list)
    return chunk_list


def _populate(array: List, begin: int = 0, end: int = -1) -> 'M3UPlaylist':
    p_list = M3UPlaylist()
    if end == -1:
        end = len(array)
    log.debug("populating playlist with rows from %s to %s", begin, end)
    entry = []
    previous_row = array[begin]
    if m3u.is_extinf_row(previous_row):
        entry.append(array[begin])
        log.debug("chunk starting with an #EXTINF row")
    for row in array[begin+1: end]:
        row = row.strip()
        log.debug("parsing row: %s", row)
        if m3u.is_extinf_row(row):
            if m3u.is_extinf_row(previous_row):
                # case of two adjacent #EXTINF rows, so a url-less entry is
                # added. This shouldn't be allowed, but sometimes those #EXTINF
                # rows are used as group separators.
                log.warning("adjacent #EXTINF rows detected")
                _append_entry(entry, p_list)
                log.debug("adding entry to the playlist: %s", entry)
                entry = []
            entry.append(row)
        elif m3u.is_comment_or_tag_row(row):
            # case of a row with a non-supported tag or a comment, so it's copied as-is
            entry.append(row)
            log.warning("commented row or unsupported tag found:\n%s", row)
        else:
            # case of a plain url row (regardless if preceded by an #EXTINF row or not)
            entry.append(row)
            log.debug("adding entry to the playlist: %s", entry)
            _append_entry(entry, p_list)
            entry = []
        previous_row = row
    return p_list


def _append_entry(entry: List, pl: M3UPlaylist):
    channel = ipytv.channel.from_playlist_entry(entry)
    pl.append_channel(channel)



import logging
import re
import urllib.parse
from typing import List

from ipytv import m3u
from ipytv.channel import IPTVChannel, IPTVAttr
from ipytv.playlist import M3UPlaylist

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class M3UDoctor:
    @staticmethod
    def _fix_split_quoted_string(m3u_rows: List) -> List:
        """
        This covers the case of rows beginning with double quotes that belong to the previous row.
        Example:
            #EXTINF:-1 tvg-id="Cinema1
            " tvg-name="Cinema1" group-title="Cinema",Cinema One
        """
        fixed_m3u_rows: List = []
        for current_row in m3u_rows:
            new_row = current_row
            if re.match(r"^\s*\"", current_row) and \
                    len(fixed_m3u_rows) > 0 and \
                    m3u.is_extinf_row(fixed_m3u_rows[-1]):
                previous_row = fixed_m3u_rows.pop()
                new_row = previous_row.rstrip() + current_row.lstrip()
            fixed_m3u_rows.append(new_row)
        return fixed_m3u_rows

    @staticmethod
    def sanitize(m3u_rows: List) -> List:
        return M3UDoctor._fix_split_quoted_string(m3u_rows)


class IPTVChannelDoctor:
    @staticmethod
    def _urlencode_value(chan: IPTVChannel, attribute_name: str) -> None:
        """
        This covers the case of tvg-logo attributes not being correctly url-encoded.
        Example (commas in the url):
            tvg-logo="https://some.image.com/images/V1_UX182_CR0,0,182,268_AL_.jpg"
        """
        if attribute_name in chan.attributes:
            value = chan.attributes[attribute_name]
            chan.attributes[attribute_name] = urllib.parse.quote(value, safe=':/%?&=')

    @staticmethod
    def _normalize_attributes_name(chan: IPTVChannel, attribute_name: str) -> None:
        """
        This covers the case of well-known attributes (i.e. the ones in IPTVAttr)
        spelled wrongly.
        Example:
            tvg-ID="" (should be tvg-id="")
        """
        try:
            IPTVAttr(attribute_name)
        except ValueError:
            try:
                key = IPTVAttr(attribute_name.lower()).value
                value = chan.attributes[attribute_name]
                del chan.attributes[attribute_name]
                chan.attributes[key] = value
            except ValueError:
                # It seems not a well-known attribute, so we leave it untouched.
                pass

    @staticmethod
    def _convert_commas(chan: IPTVChannel, attribute_name: str) -> None:
        """"
        This covers the case of attributes values containing a comma, which can confuse some
        parsing libraries (not this one, though)
        """
        if attribute_name == IPTVAttr.TVG_LOGO.value:
            return
        value: str = chan.attributes[attribute_name]
        if "," in value:
            value = value.replace(",", "_")
            chan.attributes[attribute_name] = value

    @staticmethod
    def sanitize(chan: IPTVChannel) -> IPTVChannel:
        attr: str
        new_chan = chan.copy()
        IPTVChannelDoctor._urlencode_value(new_chan, IPTVAttr.TVG_LOGO.value)
        for attr in chan.attributes.keys():
            IPTVChannelDoctor._convert_commas(new_chan, attr)
            IPTVChannelDoctor._normalize_attributes_name(new_chan, attr)
        return new_chan


class M3UPlaylistDoctor:
    @staticmethod
    def sanitize(playlist: M3UPlaylist):
        """
        This makes sure that all well-known attributes in the playlist are spelled correctly
        and that no commas appear in the attributes value.
        """
        new_playlist: M3UPlaylist = M3UPlaylist()
        chan: IPTVChannel
        for chan in playlist:
            new_playlist.append_channel(IPTVChannelDoctor.sanitize(chan))
        return new_playlist



import pathlib

from setuptools import setup, find_packages


def parse_requirements(requirements_file):
    with open(requirements_file) as req:
        return req.read().strip().split('\n')


def get_pkgdata():
    data = {}
    with open("pkgdata.txt") as f:
        lines = f.readlines()
    for row in lines:
        row_s = str(row).strip()
        if not row_s.startswith("#"):
            key = row_s.split("=")[0]
            value = row_s.split("=")[1]
            data[key] = value
    return data


here = pathlib.Path(__file__).parent.resolve()
pkgdata = get_pkgdata()
print(pkgdata)

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')
package_name = pkgdata["package.name"]
package_version = pkgdata["package.version"]


setup(
    name=package_name,
    version=package_version,
    description='A library for handling M3U playlists for IPTV (AKA m3u_plus)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Beer4Ever83/ipytv',
    author='Francesco Rainone',
    author_email='beer4evah@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Multimedia :: Video',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    license='MIT',
    keywords='m3u, m3u_plus, iptv, playlist',
    package_dir={'ipytv': 'ipytv'},
    packages=find_packages(where='.', exclude=['tests']),
    python_requires='>=3.6, <4',
    install_requires=parse_requirements("requirements.txt"),
    extras_require={},
    package_data={},
    data_files=[],
    entry_points={},
    project_urls={
        'Bug Reports': 'https://github.com/Beer4Ever83/ipytv/issues',
        'Funding': 'https://www.buymeacoffee.com/beer4ever83',
        'Source': 'https://github.com/Beer4Ever83/ipytv',
    },
)
